//-*-C++-*-

#include <andromeda/nlp/ent/numval.h>
#include <andromeda/nlp/ent/geoloc.h>

#include <andromeda/nlp/ent/name.h>
#include <andromeda/nlp/ent/link.h>
#include <andromeda/nlp/ent/cite.h>
#include <andromeda/nlp/ent/quote.h>

#include <andromeda/nlp/ent/parenthesis.h>
#include <andromeda/nlp/ent/expression.h>

#include <andromeda/nlp/ent/sentence.h>
#include <andromeda/nlp/ent/reference.h>

#include <andromeda/nlp/ent/pos_pattern.h>

#include <andromeda/nlp/ent/custom_crf.h>
#include <andromeda/nlp/ent/custom_crf/ucmi/material.h>
