//-*-C++-*-

/*
#ifndef ANDROMEDA_EXPORTERS_H_
#define ANDROMEDA_EXPORTERS_H_

#include <andromeda/tooling/structs.h>

namespace andromeda
{

  template<subject_name name>
  class exporter
  {};
  
}

#include <andromeda/tooling/exporters/paragraph.h>
#include <andromeda/tooling/exporters/webdoc.h>

#endif
*/
