//-*-C++-*-

#include <andromeda/tooling/models/base_crf_model/structures/path.h>

#include <andromeda/tooling/models/base_crf_model/structures/feature.h>
#include <andromeda/tooling/models/base_crf_model/structures/feature_bag.h>

#include <andromeda/tooling/models/base_crf_model/structures/sample.h>
#include <andromeda/tooling/models/base_crf_model/structures/sample_sequence.h>

#include <andromeda/tooling/models/base_crf_model/structures/state.h>
#include <andromeda/tooling/models/base_crf_model/structures/state_sequence.h>

#include <andromeda/tooling/models/base_crf_model/structures/string_bag.h>
#include <andromeda/tooling/models/base_crf_model/structures/string_dict.h>
#include <andromeda/tooling/models/base_crf_model/structures/string_hash.h>

#include <andromeda/tooling/models/base_crf_model/structures/token.h>
#include <andromeda/tooling/models/base_crf_model/structures/parenthesis_converter.h>

#include <andromeda/tooling/models/base_crf_model/structures/crf_model.h>
#include <andromeda/tooling/models/base_crf_model/structures/crf_model.cpp>
#include <andromeda/tooling/models/base_crf_model/structures/crf_utils.cpp>
