//-*-C++-*-

#ifndef ANDROMEDA_CRF_SAMPLE_SEQUENCE_H
#define ANDROMEDA_CRF_SAMPLE_SEQUENCE_H

namespace andromeda_crf
{
  namespace utils
  {
    struct crf_sample_sequence
    {
      std::vector<crf_sample> vs;
    };

  }

}

#endif
