//-*-C++-*-

#include <andromeda/tooling/structs/items/pos.h> // part-of-speech
#include <andromeda/tooling/structs/items/cls.h> // classification
#include <andromeda/tooling/structs/items/ent.h> // entity
#include <andromeda/tooling/structs/items/rel.h> // relation (binary)
#include <andromeda/tooling/structs/items/rec.h> // relation (multi)

