//-*-C++-*-

#ifndef PYBIND_ANDROMEDA_GLM_INTERFACE_H
#define PYBIND_ANDROMEDA_GLM_INTERFACE_H

#include <pybind/base_log.h>

#include <pybind/base_log.h>
#include <pybind/base_resources.h>

#include <pybind/glm_interface/query.h>
#include <pybind/glm_interface/model.h>

#endif

