//-*-C++-*-

#include <andromeda/tooling/models/utils.h>

#include <andromeda/tooling/models/base.h>

#include <andromeda/tooling/models/base_tok_model.h>

#include <andromeda/tooling/models/base_fst_model.h>
#include <andromeda/tooling/models/base_fst_model/fasttext_supervised_model.h>

#include <andromeda/tooling/models/base_rgx_model.h>
#include <andromeda/tooling/models/base_dct_model.h>

#include <andromeda/tooling/models/base_crf_model.h>

#include <andromeda/tooling/models/tmpl.h>
