//-*-C++-*-

#include <andromeda/tooling/producers/base/base_producer.h>
#include <andromeda/tooling/producers/base/tmpl_producer.h>

#include <andromeda/tooling/producers/impl/paragraph.h>
#include <andromeda/tooling/producers/impl/prompt.h>

#include <andromeda/tooling/producers/impl/document.h>

#include <andromeda/tooling/producers/utils.h>
